package be.iminds.ilabt.jfed.experimenter_gui.config.util;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.jfed.lowlevel.testbed_info.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.authority.finder.AuthorityFinder;
import be.iminds.ilabt.jfed.rspec.generator.RSpecGenerator;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.util.jsonld.impl.PrimaryIdObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * GuiConfigRSpecGenerator
 */
public class GuiConfigRSpecGenerator implements RSpecGenerator {
    private static final Logger LOG = LoggerFactory.getLogger(GuiConfigRSpecGenerator.class);

    @Nonnull private final JFedGuiConfig jFedGuiConfig;
    @Nonnull private final AuthorityFinder authorityFinder;
    @Nonnull private final TestbedInfoSource testbedInfoSource;

    private Server server = null;
    private Integer nodeCount = 1;
    private List<String> nodeNames = Collections.emptyList();
    private String resourceClassIdPrefix = null;
    private String nodeNamePrefix = "n";
    private List<String> nodeAnsibleGroup;
    @Nullable private String diskImage;

    @Nullable private Boolean link = null;
    @Nullable private String linkSubnet = null;
    @Nullable private Integer linkStartIpIndex = null;

    public GuiConfigRSpecGenerator(@Nonnull JFedGuiConfig jFedGuiConfig,
                                   @Nonnull AuthorityFinder authorityFinder,
                                   @Nonnull TestbedInfoSource testbedInfoSource) {
        this.jFedGuiConfig = jFedGuiConfig;
        this.authorityFinder = authorityFinder;
        this.testbedInfoSource = testbedInfoSource;
    }

    @Override
    public void setAm(@Nonnull String am) {
        if (am.startsWith("urn:")) {
            server = authorityFinder.findByAnyUrn(am, AuthorityFinder.Purpose.REQUEST_RSPEC);
        } else {
            if (am.matches("[0-9]*")) {
                server = testbedInfoSource.getServerById(Integer.parseInt(am));
            } else {
                Testbed testbed = testbedInfoSource.getTestbedById(am);
                server = testbed.getDefaultServerId() == null ?
                        null :
                        testbedInfoSource.getServerById(testbed.getDefaultServerId());
            }
        }
        if (server == null)
            throw new RuntimeException("Did not find Server for AM \""+am+"\" " +
                    "(Only URN, Testbed ID (=string) and Server ID (=int) are supported)");
    }

    @Override
    public void setNodeCount(int nodeCount) {
        this.nodeCount = nodeCount;
    }

    @Override
    public void setNodeNames(@Nonnull List<String> nodeNames) {
        this.nodeNames = nodeNames;
    }

    @Override
    public void setNodePrefix(@Nullable String prefix) {
        this.nodeNamePrefix = prefix;
    }

    @Override
    public void setResourceClassIdPrefix(@Nullable String rcIdPrefix) {
        this.resourceClassIdPrefix = rcIdPrefix;
    }

    @Override
    public void setNodeAnsibleGroup(@Nonnull List<String> nodeAnsibleGroup) {
        this.nodeAnsibleGroup = nodeAnsibleGroup;
    }

    @Override
    public void setDiskImage(@Nullable String diskImage) {
        this.diskImage = diskImage;
    }

    @Override
    public void setLink(@Nullable Boolean link) {
        this.link = link;
    }

    @Override
    public void setLinkSubnet(@Nullable String linkSubnet) {
        this.linkSubnet = linkSubnet;
    }

    @Override
    public void setLinkStartIpIndex(@Nullable Integer linkStartIpIndex) {
        this.linkStartIpIndex = linkStartIpIndex;
    }

    @Nonnull
    @Override
    public String generateRequest() {
        if (server == null)
            throw new IllegalStateException("generateRequest requires setAM(...) to set the server");

        ResourceClass chosenResourceClass;
        if (resourceClassIdPrefix == null) {
            List<ResourceClass> resourceClassList = jFedGuiConfig.findAllPossibleResourceClasses(server);
            chosenResourceClass = null;
            for (ResourceClass resourceClass : resourceClassList) {
                //highest weight is preferred
                if (chosenResourceClass == null || resourceClass.getWeightOrZero() > chosenResourceClass.getWeightOrZero()) {
                    chosenResourceClass = resourceClass;
                }
            }
            if (chosenResourceClass == null) {
                throw new RuntimeException("There is no resourceClass for server "+server.getId()+" that can be used.");
            }
            LOG.debug("Chose ResourceClass "+chosenResourceClass.getId()+" from server ResourceClasses: "+
                    resourceClassList.stream()
                            .map(PrimaryIdObject::getId)
                            .collect(Collectors.joining(", "))
            );
        } else {
            List<ResourceClass> resourceClassList = jFedGuiConfig.findAllPossibleResourceClasses(server);
            chosenResourceClass = null;
            //allow just starting with the provided ID  (this is convenient because resource classes differ in suffix for different user authorities)
            for (ResourceClass resourceClass : resourceClassList) {
                if (chosenResourceClass == null && resourceClass.getId().startsWith(resourceClassIdPrefix))
                    chosenResourceClass = resourceClass;
                if (resourceClass.getId().equalsIgnoreCase(resourceClassIdPrefix))
                    chosenResourceClass = resourceClass;
            }
            if (chosenResourceClass == null) {
                throw new RuntimeException("No resourceClass \""+resourceClassIdPrefix+"\" found. All resource classes: "+
                        jFedGuiConfig.getResourceClasses().stream().map(ResourceClass::getId).collect(Collectors.joining(", ")));
            }
            LOG.debug("Chose ResourceClass "+chosenResourceClass.getId()+" from ResourceClasses with prefix "+
                    resourceClassIdPrefix+ ": "+
                    resourceClassList.stream()
                            .map(PrimaryIdObject::getId)
                            .collect(Collectors.joining(", "))
            );
        }

        if (chosenResourceClass.getId().equals(JFedGuiConfig.RESOURCE_CLASS_ID_OTHER_SERVERS)) {
            throw new RuntimeException("There is no resource for server "+server.getId()+" " +
                    "(\""+server.getName()+"\"), cannot generate RSpec.");
        }

        Resource chosenResource = null;
        for (Resource aResource: chosenResourceClass.getResources()) {
            if (aResource.getServerId() != null && aResource.getServerId().equals(server.getId())) {
                LOG.debug("Found Resource for server: "+chosenResource);
                chosenResource = aResource;
            }
        }

        if (chosenResource == null) {
            chosenResource = jFedGuiConfig.getDefaultResource(chosenResourceClass);
            LOG.debug("Chose default Resource: "+chosenResource+"  because none matched server out of "+
                    chosenResourceClass.getResources().stream()
                            .map(PrimaryIdObject::getId)
                            .map(Object::toString)
                            .collect(Collectors.joining(", "))
            );
            throw new RuntimeException("No Resource found for ResourceClass \""+resourceClassIdPrefix+"\" and server "+server.getId()+"");
        }
        LOG.debug("Chose Resource: "+chosenResource);

        if (chosenResource.getRspecElementName() != null &&
                !chosenResource.getRspecElementName().equalsIgnoreCase("node")) {
            throw new RuntimeException("Resource for ResourceClass \""+resourceClassIdPrefix+"\" and " +
                    "server "+server.getId()+" found, but has unsupported RSpec element name: " +
                    "\""+chosenResource.getRspecElementName()+"\"");
        }

        RspecFactory factory = RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC);
        ModelRspec model = factory.createModelRspec("request");

        RspecLink rspecLink = null;
        int linkIdx = linkStartIpIndex == null ? 1 : linkStartIpIndex;
        String linkSubnetPrefix = null;
        if (Objects.equals(link, Boolean.TRUE)) {
            rspecLink = factory.createLinkWithClientId(model, "link1", "lan");
            model.addLink(rspecLink);
            if (linkSubnet != null) {
                Pattern IP_ADDRESS_PATTERN = Pattern.compile("(\\d{1,3}\\.\\d{1,3}\\.\\d{1,3})\\.(\\d{1,3})");
                Matcher matcher = IP_ADDRESS_PATTERN.matcher(linkSubnet);
                if (!matcher.matches()) {
                    throw new RuntimeException("Invalid subnet in RSpec generator config: \""+linkSubnet+"\"");
                }
                if (!matcher.group(2).equals("0")) {
                    throw new RuntimeException("Invalid subnet in RSpec generator config: \""+linkSubnet+"\" (expected to end in .0 as only subnet mask 255.255.255.0 is supported)");
                }
                linkSubnetPrefix = matcher.group(1)+".";
            } else {
                linkSubnetPrefix = "192.168.1.";
            }
        }

        for (int i = 0; i < nodeCount; i++) {
            String ansibleGroup = nodeAnsibleGroup == null ? null : (
                    i < nodeAnsibleGroup.size() ?
                            nodeAnsibleGroup.get(i) :
                            (nodeAnsibleGroup.isEmpty() ?
                                    null :
                                    nodeAnsibleGroup.get(nodeAnsibleGroup.size()-1)
                            )
            );

            RspecNode node = factory.createNode(model);
            node.setComponentManagerId(server.getDefaultComponentManagerUrn());
            node.setExclusive(chosenResource.getDefaultExclusive());

            if (rspecLink != null) {
                //this is very simple and doesn't support more than 255 IP's
                RspecInterface iface = factory.createInterface(node, rspecLink, "n" + i + ":if0");
                if (i + linkIdx >= 255) {
                    throw new RuntimeException("Too many nodes generated: ran out of IP addresses in subnet.");
                }
                iface.getIpAddresses().add(new RspecInterface.IpAddress(linkSubnetPrefix + (i + linkIdx), "255.255.255.0", "ipv4"));
            }

            if (ansibleGroup != null)
                node.getAnsibleGroups().add(ansibleGroup);

            if (i < nodeNames.size()) {
                node.setClientId(nodeNames.get(i));
            } else {
                node.setClientId(nodeNamePrefix + i);
            }

            node.setSliverTypeName(chosenResource.getSliverType());

            if (chosenResource.getHardwareType() != null) {
                node.getHardwareTypes().add(new HardwareType(chosenResource.getHardwareType()));
            }
            if (diskImage != null)
                node.setSliverTypeDiskImage(diskImage);
            else {
                if (chosenResource.getDiskImage() != null) {
                    node.setSliverTypeDiskImage(chosenResource.getDiskImage());
                }
            }

            model.addNode(node);
        }
        return model.toGeni3Rspec();
    }
}
